/*
 * HTMLCLIP.C
 *
 * Copy HTML text onto the clipboard in HTML formatting for use in applications
 * that support rich text.
 *
 * Copyright (c) 2015-2017 Malcolm J. Smith
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#define WIN32_LEAN_AND_MEAN 1
#define _CRT_SECURE_NO_WARNINGS 1
#define _CRT_NONSTDC_NO_WARNINGS 1
#include <windows.h>
#include <stdlib.h>
#include <stdio.h>

#ifdef MINICRT
#include <minicrt.h>
#endif

#include "msupd.h"

#ifndef sprintf_s
#define sprintf_s _snprintf
#endif

#ifndef FILE_SHARE_DELETE
#define FILE_SHARE_DELETE 4
#endif

const
LPSTR strUsage =
        "Usage:\n"
        "  htmlclip [filename]\n"
        "\n"
        "   -get-update    Update to latest version of current type\n"
        "   -get-daily     Get the latest daily version\n"
        "   -get-source    Get source for this version\n"
        "   -get-stable    Get the latest stable version\n"
        "   -get-symbol    Get debugging symbol for this version\n"
        "   -get-ver-x.y   Get a specific version of the program\n"
        "\n"
        " Copies either the specified file or standard input onto the clipboard as\n"
        " formatted HTML.\n";

const
CHAR strLicenseText[] = 
     "\n"
     "Copyright (c) 2015-16 Malcolm J. Smith\n"
     "\n"
     "Permission is hereby granted, free of charge, to any person obtaining a copy\n"
     "of this software and associated documentation files (the \"Software\"), to deal\n"
     "in the Software without restriction, including without limitation the rights\n"
     "to use, copy, modify, merge, publish, distribute, sublicense, and/or sell\n"
     "copies of the Software, and to permit persons to whom the Software is\n"
     "furnished to do so, subject to the following conditions:\n"
     "\n"
     "The above copyright notice and this permission notice shall be included in\n"
     "all copies or substantial portions of the Software.\n"
     "\n"
     "THE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR\n"
     "IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,\n"
     "FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE\n"
     "AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER\n"
     "LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,\n"
     "OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN\n"
     "THE SOFTWARE.\n";

BOOL Usage()
{
    ULONG BytesTransferred;
    CHAR Line[80];
    ULONG CVer;

#ifdef _MSC_VER
    CVer = _MSC_VER;
#else
    CVer = 0;
#endif

    sprintf_s(Line, sizeof(Line), "Htmlclip version %i.%i, compiled %s, %s, C version %i.%02i\n\n",
            HTMLCLIP_VER_MAJOR,
            HTMLCLIP_VER_MINOR,
            __DATE__,
#ifdef MINICRT
            "Minicrt",
#else
            "Msvcrt",
#endif
            (int)(CVer / 100),
            (int)(CVer % 100));

    WriteFile(GetStdHandle(STD_OUTPUT_HANDLE), Line, (DWORD)strlen(Line), &BytesTransferred, NULL);
    WriteFile(GetStdHandle(STD_OUTPUT_HANDLE), strUsage, (DWORD)strlen(strUsage), &BytesTransferred, NULL);
    WriteFile(GetStdHandle(STD_OUTPUT_HANDLE), strLicenseText, (DWORD)strlen(strLicenseText), &BytesTransferred, NULL);
    return TRUE;
}

typedef enum {
    HtmlClipUpdateTypeNone = 0,
    HtmlClipUpdateTypeBinary,
    HtmlClipUpdateTypeSource,
    HtmlClipUpdateTypeSymbol,
    HtmlClipUpdateTypeStable,
    HtmlClipUpdateTypeDaily
} HTMLCLIP_UPDATE_TYPE;

BOOL
HtmlClipAssembleUrl(
    LPSTR szUrl,
    DWORD szUrlLength,
    LPSTR szVerString,
    LPSTR szFileName,
    LPSTR szFileExtension
    )
{
    CHAR szTempFileName[128];

    if (szFileName == NULL) {
        sprintf_s(szTempFileName,
                  sizeof(szTempFileName)/sizeof(szTempFileName[0]),
                  "%s/%s",
#ifdef _M_AMD64
                  "amd64",
#elif defined(_M_IX86)
                  "win32",
#endif
                  "htmlclip"
                  );

        szFileName = szTempFileName;
    }

    sprintf_s(szUrl,
              szUrlLength,
              "http://www.malsmith.net/download/?obj=%s/%s/%s%s",
              "htmlclip",
              szVerString?szVerString:
#if HTMLCLIP_BUILD_ID
              "latest-daily",
#else
              "latest-stable",
#endif
              szFileName,
              szFileExtension?szFileExtension:".exe"
              );

    return TRUE;
}


BOOL
HtmlClipUpdateBinary(
    HTMLCLIP_UPDATE_TYPE UpdateType,
    LPSTR VersionToUpdateTo
    )
{
    CHAR szAgent[128];
    CHAR szDisplay[300];
    CHAR szUrl[250];
    CHAR szVer[32];
    DWORD BytesTransferred;
    LPSTR TargetName = NULL;
    UpdError Error = 0;

#if HTMLCLIP_BUILD_ID
    sprintf_s(szVer, sizeof(szVer)/sizeof(szVer[0]), "%i.%i.%i", HTMLCLIP_VER_MAJOR, HTMLCLIP_VER_MINOR, HTMLCLIP_BUILD_ID);
#else
    sprintf_s(szVer, sizeof(szVer)/sizeof(szVer[0]), "%i.%i", HTMLCLIP_VER_MAJOR, HTMLCLIP_VER_MINOR);
#endif

    sprintf_s(szAgent, sizeof(szAgent)/sizeof(szAgent[0]), "htmlclip %s", szVer);

    if (UpdateType == HtmlClipUpdateTypeBinary) {
        HtmlClipAssembleUrl(szUrl, sizeof(szUrl)/sizeof(szUrl[0]), VersionToUpdateTo, NULL, NULL);
    } else if (UpdateType == HtmlClipUpdateTypeSource) {
        HtmlClipAssembleUrl(szUrl, sizeof(szUrl)/sizeof(szUrl[0]), szVer, "htmlclip-source", ".zip");
        TargetName = "htmlclip-source.zip";
    } else if (UpdateType == HtmlClipUpdateTypeSymbol) {
        HtmlClipAssembleUrl(szUrl, sizeof(szUrl)/sizeof(szUrl[0]), szVer, NULL, ".pdb");
        TargetName = "htmlclip.pdb";
    } else if (UpdateType == HtmlClipUpdateTypeStable) {
        HtmlClipAssembleUrl(szUrl, sizeof(szUrl)/sizeof(szUrl[0]), "latest-stable", NULL, NULL);
    } else if (UpdateType == HtmlClipUpdateTypeDaily) {
        HtmlClipAssembleUrl(szUrl, sizeof(szUrl)/sizeof(szUrl[0]), "latest-daily", NULL, NULL);
    }

    sprintf_s(szDisplay, sizeof(szDisplay)/sizeof(szDisplay[0]), "Updating from %s\n", szUrl);
    WriteFile(GetStdHandle(STD_OUTPUT_HANDLE), szDisplay, (DWORD)strlen(szDisplay), &BytesTransferred, NULL);
    Error = UpdateBinaryFromUrl(szUrl, TargetName, szAgent);

    sprintf_s(szDisplay, sizeof(szDisplay)/sizeof(szDisplay[0]), "Update result: %s\n", UpdateErrorString(Error));
    WriteFile(GetStdHandle(STD_OUTPUT_HANDLE), szDisplay, (DWORD)strlen(szDisplay), &BytesTransferred, NULL);
    return TRUE;
}

//
//  Dummy header.  0.9 is part of the protocol.  The rest contain strings
//  that are 8 characters long that can be overwritten with real zero padded
//  strings.  The zero padding helps keep computation sane.
//

const char DummyHeader[] = 
                  "Version:0.9\n"
                  "StartHTML:12345678\n"
                  "EndHTML:12345678\n"
                  "StartFragment:12345678\n"
                  "EndFragment:12345678\n";

#define HTMLCLIP_HDR_SIZE (sizeof(DummyHeader)-1)

const char DummyFragStart[] = 
                  "<!--StartFragment-->";

#define HTMLCLIP_FRAGSTART_SIZE (sizeof(DummyFragStart)-1)

const char DummyFragEnd[] = 
                  "<!--EndFragment-->";

#define HTMLCLIP_FRAGEND_SIZE (sizeof(DummyFragEnd)-1)

#define MAX_PIPE_SIZE (4*1024*1024)

int __cdecl
main (
    int argc,
    char* argv[]
    )
{
    HANDLE hFile = NULL;
    DWORD  FileSize = 0;
    DWORD  AllocSize;
    HANDLE hMem;
    PUCHAR pMem;
    ULONG  BytesTransferred;
    ULONG  CurrentOffset;
    UINT   ClipFmt;
    
    LONG   CurrentArg;
    HTMLCLIP_UPDATE_TYPE UpdateType = HtmlClipUpdateTypeNone;
    LPSTR  VersionToUpdateTo = NULL;

    //
    //  If the user specified a file, go read it.  If not, read
    //  from standard input.
    //

    if (argc > 1) {

        for (CurrentArg = 1; CurrentArg < argc; CurrentArg++) {
            if (argv[CurrentArg][0] == '-' || argv[CurrentArg][0] == '/') {
                LPSTR Arg = &argv[CurrentArg][1];
                BOOL ArgParsed = FALSE;
    
                if (stricmp(Arg, "get-daily") == 0) {
                    ArgParsed = TRUE;
                    UpdateType = HtmlClipUpdateTypeDaily;
                } else if (stricmp(Arg, "get-source") == 0) {
                    ArgParsed = TRUE;
                    UpdateType = HtmlClipUpdateTypeSource;
                } else if (stricmp(Arg, "get-stable") == 0) {
                    ArgParsed = TRUE;
                    UpdateType = HtmlClipUpdateTypeStable;
                } else if (stricmp(Arg, "get-symbol") == 0) {
                    ArgParsed = TRUE;
                    UpdateType = HtmlClipUpdateTypeSymbol;
                } else if (stricmp(Arg, "get-update") == 0) {
                    ArgParsed = TRUE;
                    UpdateType = HtmlClipUpdateTypeBinary;
                } else if (strnicmp(Arg, "get-ver-", 8) == 0) {
                    ArgParsed = TRUE;
                    UpdateType = HtmlClipUpdateTypeBinary;
                    VersionToUpdateTo = &Arg[8];
                }
    
                if (!ArgParsed) {
                    Usage();
                    return EXIT_FAILURE;
                }

                if (UpdateType != HtmlClipUpdateTypeNone) {
                    HtmlClipUpdateBinary(UpdateType, VersionToUpdateTo);
                    return EXIT_SUCCESS;
                }

            } else if (hFile == NULL) {
    
                hFile = CreateFile(argv[CurrentArg], GENERIC_READ, FILE_SHARE_READ|FILE_SHARE_DELETE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL,NULL);
                if (hFile == INVALID_HANDLE_VALUE) {
                    printf("Could not open file, error %i\n", (int)GetLastError());
                    return EXIT_FAILURE;
                }
    
                FileSize = GetFileSize(hFile, NULL);
            }
        }

    } else {
        DWORD FileType;

        FileSize = MAX_PIPE_SIZE;
        hFile = GetStdHandle(STD_INPUT_HANDLE);

        //
        //  As a bit of a hack, if the input is stdin and stdin is to
        //  a console (as opposed to a pipe or file) assume the user isn't
        //  sure how to run this program and help them along.
        //

        FileType = GetFileType(hFile);
        FileType = FileType & ~(FILE_TYPE_REMOTE);
        if (FileType == FILE_TYPE_CHAR) {
            Usage();
            return EXIT_FAILURE;
        }
    }

    //
    //  Allocate space to copy the file or pipe contents.
    //

    AllocSize = FileSize + HTMLCLIP_HDR_SIZE + HTMLCLIP_FRAGSTART_SIZE + HTMLCLIP_FRAGEND_SIZE + 2;
    hMem = GlobalAlloc(GMEM_MOVEABLE|GMEM_DDESHARE, AllocSize);
    if (hMem == NULL) {
        printf("Could not allocate memory, error %i\n", (int)GetLastError());
        return EXIT_FAILURE;
    }

    pMem = GlobalLock(hMem);

    //
    //  Note this is not Unicode.
    //
    //  First, prepare the header.  Leave a magic space at the end of the
    //  header.
    //

    sprintf_s((PCHAR)pMem,
              AllocSize,
              "Version:0.9\n"
              "StartHTML:%08i\n"
              "EndHTML:%08i\n"
              "StartFragment:%08i\n"
              "EndFragment:%08i\n"
              "<!--StartFragment-->"
              " ",
              (int)HTMLCLIP_HDR_SIZE,
              (int)(HTMLCLIP_HDR_SIZE + HTMLCLIP_FRAGSTART_SIZE + HTMLCLIP_FRAGEND_SIZE + 1 + FileSize),
              (int)(HTMLCLIP_HDR_SIZE + HTMLCLIP_FRAGSTART_SIZE),
              (int)(HTMLCLIP_HDR_SIZE + HTMLCLIP_FRAGSTART_SIZE + 1 + FileSize));

    CurrentOffset = 0;

    //
    //  Read text immediately following the header that we just prepared.
    //

    while (ReadFile(hFile, pMem + HTMLCLIP_HDR_SIZE + HTMLCLIP_FRAGSTART_SIZE + 1 + CurrentOffset, FileSize - CurrentOffset, &BytesTransferred, NULL)) {
        CurrentOffset += BytesTransferred;

        //
        //  If we're reading from the pipe and have exceeded our storage,
        //  fail.
        //

        if (CurrentOffset == FileSize &&
            FileSize == MAX_PIPE_SIZE &&
            hFile == GetStdHandle(STD_INPUT_HANDLE)) {

            printf("Pipe data too large for buffer.  Limit is %i bytes\n", MAX_PIPE_SIZE);
            return EXIT_FAILURE;
        }

        if (BytesTransferred == 0) {
            break;
        }
    }

    if (CurrentOffset == 0) {
        Usage();
        return EXIT_FAILURE;
    }

    //
    //  If we don't have as much text as we expected, the header is wrong
    //  and needs to be rewritten with correct values.
    //

    if (CurrentOffset < FileSize) {
        FileSize = CurrentOffset;
        sprintf_s((PCHAR)pMem,
                  AllocSize,
                  "Version:0.9\n"
                  "StartHTML:%08i\n"
                  "EndHTML:%08i\n"
                  "StartFragment:%08i\n"
                  "EndFragment:%08i\n"
                  "<!--StartFragment-->",
               (int)HTMLCLIP_HDR_SIZE,
               (int)(HTMLCLIP_HDR_SIZE + HTMLCLIP_FRAGSTART_SIZE + HTMLCLIP_FRAGEND_SIZE + 1 + FileSize),
               (int)(HTMLCLIP_HDR_SIZE + HTMLCLIP_FRAGSTART_SIZE),
               (int)(HTMLCLIP_HDR_SIZE + HTMLCLIP_FRAGSTART_SIZE + 1 + FileSize));

        //
        //  That magic space from above has now been set to NULL by printf.
        //  Put the magic space back.
        //

        pMem[HTMLCLIP_HDR_SIZE + HTMLCLIP_FRAGSTART_SIZE] = ' ';
    }

    //
    //  Fill in the footer of the protocol.
    //

    sprintf_s((PCHAR)(pMem + FileSize + HTMLCLIP_HDR_SIZE + HTMLCLIP_FRAGSTART_SIZE + 1),
              sizeof(DummyFragEnd),
              "%s",
              DummyFragEnd);

    GlobalUnlock(hMem);

    //
    //  Send the buffer to the clipboard.
    //

    if (!OpenClipboard(NULL)) {
        printf("Could not open clipboard, error %i\n", (int)GetLastError());
        return EXIT_FAILURE;
    }

    if (!EmptyClipboard()) {
        printf("Could not empty clipboard, error %i\n", (int)GetLastError());
        return EXIT_FAILURE;
    }

    ClipFmt = RegisterClipboardFormat("HTML Format");
    if (ClipFmt == 0) {
        printf("Could not register clipboard format, error %i\n", (int)GetLastError());
        return EXIT_FAILURE;
    }

    if (SetClipboardData(ClipFmt, hMem) == NULL) {
        printf("Could not set clipboard data, error %i\n", (int)GetLastError());
        return EXIT_FAILURE;
    }

    GlobalFree(hMem);
    if (argc > 1) {
        CloseHandle(hFile);
    }
    return TRUE;
}

// vim:sw=4:ts=4:et:
